from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import simplejson

from amaascore.config import ENDPOINTS
from amaascore.core.interface import Interface
from amaascore.transactions.utils import json_to_transaction, json_to_position


class TransactionsInterface(Interface):

    def __init__(self, logger=None):
        endpoint = ENDPOINTS.get('transactions')
        self.logger = logger or logging.getLogger(__name__)
        self.json_header = {'Content-Type': 'application/json'}
        super(TransactionsInterface, self).__init__(endpoint=endpoint)

    def new(self, transaction):
        url = self.endpoint + '/transactions'
        response = self.session.post(url, json=transaction.to_interface())
        if response.ok:
            transaction = json_to_transaction(response.json())
            return transaction
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def amend(self, transaction):
        url = '%s/transactions/%s/%s' % (self.endpoint, transaction.asset_manager_id, transaction.transaction_id)
        response = self.session.put(url, json=transaction.to_interface())
        if response.ok:
            transaction = json_to_transaction(response.json())
            return transaction
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve(self, asset_manager_id, transaction_id):
        url = '%s/transactions/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        response = self.session.get(url)
        if response.ok:
            return json_to_transaction(response.json())
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def transactions_by_asset_manager(self, asset_manager_id):
        url = '%s/transactions/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url)
        if response.ok:
            transactions = [json_to_transaction(json_transaction) for json_transaction in response.json()]
            return transactions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def cancel(self, asset_manager_id, transaction_id):
        url = '%s/transactions/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        response = self.session.delete(url)
        if response.ok:
            print("DO SOMETHING?")
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def search(self, asset_manager_ids=None, transaction_ids=None):
        search_params = {}
        # Potentially roll this into a loop through args rather than explicitly named - depends on additional validation
        if asset_manager_ids:
            search_params['asset_manager_ids'] = asset_manager_ids
        if transaction_ids:
            search_params['transaction_ids'] = transaction_ids
        url = self.endpoint + '/transactions'
        response = self.session.get(url, params=search_params)
        if response.ok:
            transactions = [json_to_transaction(json_transaction) for json_transaction in response.json()]
            return transactions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def position_search(self, asset_manager_ids=None, book_ids=None, account_ids=None, accounting_types=None,
                        asset_ids=None, position_date=None):
        url = self.endpoint + '/positions'
        search_params = {}
        # Potentially roll into a loop
        if asset_manager_ids:
            search_params['asset_manager_ids'] = asset_manager_ids
        if book_ids:
            search_params['book_ids'] = book_ids
        if account_ids:
            search_params['account_ids'] = account_ids
        if accounting_types:
            search_params['accounting_types'] = accounting_types
        if asset_ids:
            search_params['asset_ids'] = asset_ids
        if position_date:
            search_params['position_date'] = position_date
        response = self.session.get(url, params=search_params)
        if response.ok:
            positions = [json_to_position(json_position) for json_position in response.json()]
            return positions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def positions_by_asset_manager_book(self, asset_manager_id, book_id):
        url = '%s/positions/%s/%s' % (self.endpoint, asset_manager_id, book_id)
        response = self.session.get(url)
        if response.ok:
            positions = [json_to_position(json_position) for json_position in response.json()]
            return positions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def positions_by_asset_manager(self, asset_manager_id):
        url = '%s/positions/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url)
        if response.ok:
            positions = [json_to_position(json_position) for json_position in response.json()]
            return positions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def allocate_transaction(self, asset_manager_id, transaction_id, allocation_type, allocation_dicts):
        """

        :param asset_manager_id:
        :param transaction_id:
        :param allocation_type:
        :param allocation_dicts:
        :return:
        """
        url = '%s/allocations/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        params = {'allocation_type': allocation_type}
        # As per https://github.com/kennethreitz/requests/issues/2755 - requests doesn't support custom Encoders, and
        # the default encoding fails on Decimals in Python 3 (but not in Python 2?)
        response = self.session.post(url, params=params, data=simplejson.dumps(allocation_dicts),
                                     headers=self.json_header)
        if response.ok:
            allocations = [json_to_transaction(json_allocation) for json_allocation in response.json()]
            return allocations
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve_transaction_allocations(self, asset_manager_id, transaction_id):
        """

        :param asset_manager_id:
        :param transaction_id:
        :return:
        """
        url = '%s/allocations/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        response = self.session.get(url)
        if response.ok:
            allocations = [json_to_transaction(json_allocation) for json_allocation in response.json()]
            return allocations
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def upsert_transaction_asset(self, transaction_asset_json):
        """
        This API should not be called in normal circumstances as the asset cache will populate itself from the assets
        which are created via the Assets API.  However, it can be useful for certain testing scenarios.
        :param transaction_asset_json:
        :return:
        """
        url = self.endpoint + '/assets'
        response = self.session.post(url, json=transaction_asset_json)
        if response.ok:
            return response.json()
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def upsert_transaction_book(self, transaction_book_json):
        """
        This API should not be called in normal circumstances as the book cache will populate itself from the book
        which are created via the Books API.  However, it can be useful for certain testing scenarios.
        :param transaction_book_json:
        :return:
        """
        url = self.endpoint + '/books'
        response = self.session.post(url, json=transaction_book_json)
        if response.ok:
            return response.json()
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def net_transactions(self, asset_manager_id, transaction_ids, netting_type='Net'):
        """

        :param asset_manager_id: The asset_manager_id of the netting set owner
        :param transaction_ids:  A list of transaction_ids to net
        :param netting_type:
        :return:
        """
        url = '%s/netting/%s' % (self.endpoint, asset_manager_id)
        params = {'netting_type': netting_type}
        response = self.session.post(url, params=params, json=transaction_ids)
        if response.ok:
            net_transaction = json_to_transaction(response.json())
            return net_transaction
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve_netting_set(self, asset_manager_id, transaction_id):
        """
        Returns all the transaction_ids associated with a single netting set.  Pass in the ID for any transaction in
        the set.
        :param asset_manager_id:  The asset_manager_id for the netting set owner.
        :param transaction_id: A transaction_id of an entry within the netting set.
        :return:
        """
        url = '%s/netting/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        response = self.session.get(url)
        if response.ok:
            net_transaction_id, netting_set_json = next(iter(response.json().items()))
            return net_transaction_id, [json_to_transaction(net_transaction) for net_transaction in netting_set_json]
        else:
            self.logger.error(response.text)
            response.raise_for_status()



from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import datetime, date
from dateutil import parser

from amaascore.assets.real_asset import RealAsset


class RealEstate(RealAsset):

    def __init__(self, asset_manager_id, asset_id, asset_issuer_id=None, asset_status='Active', description='',
                 country_id=None, venue_id=None, currency=None, links={}, references={}, *args, **kwargs):
        super(RealEstate, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id,
                                         asset_issuer_id=asset_issuer_id, asset_status=asset_status,
                                         description=description, country_id=country_id, venue_id=venue_id,
                                         currency=currency, links=links, references=references,
                                         *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import datetime, date
from dateutil import parser

from amaascore.assets.fund import Fund


class ExchangeTradedFund(Fund):

    def __init__(self, asset_manager_id, asset_id, asset_issuer_id=None, asset_status='Active', description='',
                 country_id=None, venue_id=None, currency=None, issue_date=None, links={}, references={},
                 *args, **kwargs):
        self.issue_date = issue_date
        super(ExchangeTradedFund, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id,
                                                 fungible=True, asset_issuer_id=asset_issuer_id,
                                                 asset_status=asset_status, description=description,
                                                 country_id=country_id, venue_id=venue_id,
                                                 currency=currency, links=links, references=references,
                                                 *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import date, datetime
from dateutil import parser
from decimal import Decimal

from amaascore.assets.asset import Asset


class ListedDerivative(Asset):

    def __init__(self, asset_manager_id, asset_id, maturity_date=date.max, asset_issuer_id=None, asset_status='Active',
                 description='', country_id=None, venue_id=None, issue_date=date.min, links={}, references={},
                 *args, **kwargs):
        if not hasattr(self, 'asset_class'):  # A more specific child class may have already set this
            self.asset_class = 'ListedDerivative'

        super(ListedDerivative, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id, fungible=True,
                                               asset_issuer_id=asset_issuer_id, asset_status=asset_status,
                                               description=description, country_id=country_id, venue_id=venue_id,
                                               issue_date=issue_date, maturity_date=maturity_date,
                                               links=links, references=references,
                                               *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.parties.company import Company


class Exchange(Company):

    def __init__(self, asset_manager_id, party_id, base_currency=None, description='', party_status='Active',
                 addresses={}, emails={}, links={}, references={}, *args, **kwargs):
        super(Exchange, self).__init__(asset_manager_id=asset_manager_id, party_id=party_id,
                                       base_currency=base_currency, description=description,
                                       party_status=party_status,
                                       addresses=addresses, emails=emails, links=links, references=references,
                                       *args, **kwargs)

# TODO - add exchange market open hours



from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import date
from decimal import Decimal

from amaascore.assets.future import Future


class BondFuture(Future):

    def __init__(self, asset_manager_id, asset_id, expiry_date, underlying_bond_tenor, underlying_bond_coupon,
                 cheapest_to_deliver_id=None, asset_issuer_id=None, asset_status='Active', issue_date=date.min,
                 description='', country_id=None, venue_id=None, links={}, references={}, *args, **kwargs):
        """

        :param asset_manager_id: The asset manager who owns the data for this BondFuture
        :param asset_id: The owning asset manager's unique identifier for this BondFuture
        :param expiry_date: The date on which this BondFuture will expire
        :param underlying_bond_tenor: The tenor of the bond on which this is a future
        :param cheapest_to_deliver_id: Populated from the server, cannot be calculated in the SDK.
        :param asset_issuer_id: The owning asset manager's identifier for the issuer of this future
        :param asset_status: The status of this BondFuture
        :param issue_date: The date on which this BondFuture was created
        :param description: A human-readable description of this BondFuture
        :param country_id: The ISO 3166-1 alpha-3 code for the country that this BondFutureOption was issued in
        :param venue_id: The ISO-10383 identifier for the venue on which this BondFutureOption trades
        :param references: A dictionary of additional References for this BondFutureOption
        :param args:
        :param kwargs:
        """
        self.cheapest_to_deliver_id = cheapest_to_deliver_id
        self.expiry_date = expiry_date
        self.underlying_bond_tenor = underlying_bond_tenor
        self.underlying_bond_coupon = underlying_bond_coupon
        super(BondFuture, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id,
                                         asset_issuer_id=asset_issuer_id, asset_status=asset_status,
                                         description=description, country_id=country_id, venue_id=venue_id,
                                         links=links, references=references, issue_date=issue_date,
                                         *args, **kwargs)

    @property
    def underlying_bond_coupon(self):
        if hasattr(self, '_underlying_bond_coupon'):
            return self._underlying_bond_coupon

    @underlying_bond_coupon.setter
    def underlying_bond_coupon(self, underlying_bond_coupon):
        """
        The coupon paid out by the underlying bond.  Represented as a fraction of 1 (e.g. 0.05 is 5%).
        :param underlying_bond_coupon:
        :return:
        """
        if underlying_bond_coupon is not None:
            self._underlying_bond_coupon = Decimal(underlying_bond_coupon)




from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.asset_managers.asset_manager import AssetManager


def json_to_asset_manager(json_asset_manager):
    asset_manager = AssetManager(**json_asset_manager)
    return asset_manager



from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import datetime, date
from dateutil import parser
import sys

from amaascore.assets.asset import Asset

# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


class Fund(Asset):

    def __init__(self, asset_manager_id, asset_id, asset_issuer_id=None, asset_status='Active', description='',
                 country_id=None, venue_id=None, currency=None, creation_date=None,
                 links={}, references={},
                 *args, **kwargs):
        if not hasattr(self, 'asset_class'):  # A more specific child class may have already set this
            self.asset_class = 'Fund'
        self.creation_date = creation_date
        super(Fund, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id, fungible=True,
                                   asset_issuer_id=asset_issuer_id, asset_status=asset_status,
                                   description=description, country_id=country_id, venue_id=venue_id,
                                   currency=currency, links=links, references=references, *args, **kwargs)

    #  Perhaps this comes from the linked Fund
    @property
    def creation_date(self):
        if hasattr(self, '_creation_date'):
            return self._creation_date

    @creation_date.setter
    def creation_date(self, value):
        """
        The date on which the bond was issued.
        :param creation_date:
        :return:
        """
        if value:
            self._creation_date = parse(value).date() if isinstance(value, type_check) else value



from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import re

from amaascore.error_messages import ERROR_LOOKUP
from amaascore.core.amaas_model import AMaaSModel
from amaascore.core.reference import Reference
from amaascore.parties.children import Address, Email
from amaascore.parties.enums import PARTY_STATUSES


class Party(AMaaSModel):

    @staticmethod
    def children():
        """ A dict of which of the attributes are collections of other objects, and what type """
        return {'addresses': Address, 'emails': Email, 'references': Reference}

    def __init__(self, asset_manager_id, party_id, party_status='Active', base_currency=None, description='',
                 addresses={}, emails={}, links={}, references={}, *args, **kwargs):
        self.asset_manager_id = asset_manager_id
        self.party_id = party_id
        self.party_status = party_status
        if not hasattr(self, 'party_class'):  # A more specific child class may have already set this
            self.party_class = 'Party'
        self.party_type = self.__class__.__name__
        self.base_currency = base_currency
        self.description = description
        self.addresses = addresses
        self.emails = emails
        self.links = links
        self.references = references
        super(Party, self).__init__(*args, **kwargs)

    def upsert_address(self, address_type, address):
        addresses = copy.deepcopy(self.addresses)
        addresses.update({address_type: address})
        self.addresses = addresses

    @property
    def addresses(self):
        if hasattr(self, '_addresses'):
            return self._addresses

    @addresses.setter
    def addresses(self, addresses):
        if not isinstance(addresses, dict):
            raise ValueError(ERROR_LOOKUP.get('address_invalid') % (str(addresses), self.party_id,
                                                                    self.asset_manager_id))
        primary = [address.address_primary for address in addresses.values() if address.address_primary]
        # If addresses are present, one of them must be primary
        if len(addresses) and len(primary) != 1:
            raise ValueError(ERROR_LOOKUP.get('address_primary') % (self.party_id, self.asset_manager_id))
        self._addresses = addresses

    def upsert_email(self, email_type, email):
        emails = copy.deepcopy(self.emails)
        emails.update({email_type: email})
        self.emails = emails

    @property
    def emails(self):
        if hasattr(self, '_emails'):
            return self._emails

    @emails.setter
    def emails(self, emails):
        if not isinstance(emails, dict):
            raise ValueError(ERROR_LOOKUP.get('email_invalid') % (str(emails), self.party_id, self.asset_manager_id))
        primary = [email.email_primary for email in emails.values() if email.email_primary]
        # If emails are present, one of them must be primary
        if len(emails) and len(primary) != 1:
            raise ValueError(ERROR_LOOKUP.get('email_primary') % (self.party_id, self.asset_manager_id))
        # Validate email addresses
        invalid = [email.email for email in emails.values() if not re.match('[^@]+@[^@]+\.[^@]+', email.email)]
        if invalid:
            raise ValueError(ERROR_LOOKUP.get('email_address_invalid') % (str(invalid), self.party_id,
                                                                          self.asset_manager_id))
        self._emails = emails

    @property
    def party_status(self):
        if hasattr(self, '_party_status'):
            return self._party_status

    @party_status.setter
    def party_status(self, party_status):
        """

        :param transaction_status: The status of the transaction - e.g. Active, Inactive
        :return:
        """
        if party_status not in PARTY_STATUSES:
            raise ValueError(ERROR_LOOKUP.get('party_status_invalid') % (party_status, self.party_id,
                                                                         self.asset_manager_id))
        else:
            self._party_status = party_status



from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import date, datetime
from dateutil import parser
from decimal import Decimal

from amaascore.assets.derivative import Derivative
from amaascore.assets.option_mixin import OptionMixin


class BondOption(Derivative, OptionMixin):

    def __init__(self, asset_manager_id, option_type, strike, underlying_asset_id, option_style, asset_id=None,
                 asset_issuer_id=None, asset_status='Active', description='', country_id=None, venue_id=None,
                 issue_date=date.min, expiry_date=date.max, links={}, references={}, *args, **kwargs):
        self.option_type = option_type
        self.strike = strike
        self.underlying_asset_id = underlying_asset_id
        self.option_style = option_style
        super(BondOption, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id,
                                         asset_issuer_id=asset_issuer_id, asset_status=asset_status,
                                         description=description, country_id=country_id, venue_id=venue_id,
                                         expiry_date=expiry_date, links=links, references=references,
                                         issue_date=issue_date, *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

from configparser import ConfigParser, NoSectionError
from os.path import expanduser, join
import requests

from amaascore.exceptions import AMaaSException


class Interface(object):
    """
    Currently this class doesn't do anything - but I anticipate it will be needed in the future.
    """

    def __init__(self, endpoint, use_auth=False, config_filename=None):
        # Add authentication routine here
        self.config_filename = config_filename
        self.auth_token = self.read_token() if use_auth is True else ''
        self.endpoint = endpoint
        self.session = requests.Session()
        self.session.headers.update({'Authorization': self.auth_token})

    @staticmethod
    def generate_config_filename():
        home = expanduser("~")
        return join(home, '.amaas.cfg')

    def read_token(self):
        if self.config_filename is None:
            self.config_filename = self.generate_config_filename()
        parser = ConfigParser()
        parser.read(self.config_filename)
        try:
            token = parser.get(section='auth', option='token')
        except NoSectionError:
            raise AMaaSException('Invalid AMaaS config file')
        return token



from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import datetime
from dateutil.parser import parse
from decimal import Decimal
import sys
import uuid

from amaascore.error_messages import ERROR_LOOKUP
from amaascore.exceptions import TransactionNeedsSaving
from amaascore.core.amaas_model import AMaaSModel
from amaascore.core.reference import Reference
from amaascore.transactions.children import Charge, Code, Comment, Link, Party
from amaascore.transactions.enums import TRANSACTION_ACTIONS, TRANSACTION_STATUSES, TRANSACTION_TYPES

# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


class Transaction(AMaaSModel):

    @staticmethod
    def children():
        """ A dict of which of the attributes are collections of other objects, and what type """
        return {'charges': Charge, 'codes': Code, 'comments': Comment, 'links': Link, 'parties': Party,
                'references': Reference}

    def __init__(self, asset_manager_id, asset_book_id, counterparty_book_id, transaction_action, asset_id, quantity,
                 transaction_date, settlement_date, price, transaction_currency, settlement_currency=None,
                 asset=None, execution_time=None, transaction_type='Trade', transaction_id=None,
                 transaction_status='New', charges=None, codes=None, comments=None, links=None, parties=None,
                 references=None, *args, **kwargs):
        """

        :param asset_manager_id:
        :param asset_book_id:
        :param counterparty_book_id:
        :param transaction_action:
        :param asset_id:
        :param quantity:
        :param transaction_date:
        :param settlement_date:
        :param price:
        :param transaction_currency:
        :param settlement_currency: The currency in which the transaction will be settled.  Defaults to the
        transaction_currency if not specified.
        :param asset:
        :param execution_time:
        :param transaction_type:
        :param transaction_id:
        :param transaction_status:
        :param charges:
        :param codes:
        :param comments:
        :param links:
        :param parties:
        :param references:
        :param args:
        :param kwargs:
        """

        self.transaction_id = transaction_id or uuid.uuid4().hex
        self.asset_manager_id = asset_manager_id
        self.asset_book_id = asset_book_id
        self.counterparty_book_id = counterparty_book_id
        self.transaction_action = transaction_action
        self.asset_id = asset_id  # This is duplicated on the child asset.  Remove?
        self.quantity = quantity
        self.transaction_date = transaction_date
        self.settlement_date = settlement_date
        self.price = price
        self.transaction_currency = transaction_currency
        self.settlement_currency = settlement_currency or transaction_currency
        self.transaction_type = transaction_type
        self.transaction_status = transaction_status

        # Cannot be in method signature or the value gets bound to the constructor call
        self.execution_time = execution_time or datetime.datetime.utcnow()

        self.charges = charges or {}
        self.codes = codes or {}
        self.comments = comments or {}
        self.links = links or {}
        self.parties = parties or {}
        self.references = references or {}
        self.references['AMaaS'] = Reference(reference_value=self.transaction_id)  # Upserts the AMaaS Reference

        self.postings = []
        self.asset = asset
        super(Transaction, self).__init__(*args, **kwargs)

    @property
    def quantity(self):
        return self._quantity

    @quantity.setter
    def quantity(self, value):
        """
        Force the quantity to always be a decimal
        :param value:
        :return:
        """
        self._quantity = Decimal(value)

    @property
    def price(self):
        return self._price

    @price.setter
    def price(self, value):
        """
        Force the price to always be a decimal
        :param value:
        :return:
        """
        self._price = Decimal(value)

    @property
    def transaction_date(self):
        return self._transaction_date

    @transaction_date.setter
    def transaction_date(self, value):
        """
        Force the transaction_date to always be a date
        :param value:
        :return:
        """
        if value:
            self._transaction_date = parse(value).date() if isinstance(value, type_check) else value

    @property
    def settlement_date(self):
        return self._settlement_date

    @settlement_date.setter
    def settlement_date(self, value):
        """
        Force the settlement_date to always be a date
        :param value:
        :return:
        """
        if value:
            self._settlement_date = parse(value).date() if isinstance(value, type_check) else value

    @property
    def execution_time(self):
        return self._execution_time

    @execution_time.setter
    def execution_time(self, value):
        """
        Force the execution_time to always be a datetime
        :param value:
        :return:
        """
        if value:
            self._execution_time = parse(value) if isinstance(value, type_check) else value

    @property
    def gross_settlement(self):
        if hasattr(self, '_gross_settlement'):
            return self.__gross_settlement
        return self.quantity * self.price

    @gross_settlement.setter
    def gross_settlement(self, gross_settlement):
        """

        :param gross_settlement:
        :return:
        """
        if gross_settlement:
            self._gross_settlement = Decimal(gross_settlement)

    @property
    def net_settlement(self):
        if hasattr(self, '_net_settlement'):
            return self._net_settlement
        return self.gross_settlement - self.charges_net_effect()

    @net_settlement.setter
    def net_settlement(self, net_settlement):
        """

        :param gross_settlement:
        :return:
        """
        if net_settlement:
            self._net_settlement = Decimal(net_settlement)

    @property
    def transaction_action(self):
        if hasattr(self, '_transaction_action'):
            return self._transaction_action

    @transaction_action.setter
    def transaction_action(self, transaction_action):
        """

        :param transaction_action: The action that this transaction is recording - e.g. Buy, Deliver
        :return:
        """
        if transaction_action not in TRANSACTION_ACTIONS:
            raise ValueError(ERROR_LOOKUP.get('transaction_action_invalid') % (transaction_action, self.transaction_id,
                                                                               self.asset_manager_id))
        else:
            self._transaction_action = transaction_action

    @property
    def transaction_status(self):
        if hasattr(self, '_transaction_status'):
            return self._transaction_status

    @transaction_status.setter
    def transaction_status(self, transaction_status):
        """

        :param transaction_status: The status of the transaction - e.g. New, Netted
        :return:
        """
        if transaction_status not in TRANSACTION_STATUSES:
            raise ValueError(ERROR_LOOKUP.get('transaction_status_invalid') % (transaction_status, self.transaction_id,
                                                                               self.asset_manager_id))
        else:
            self._transaction_status = transaction_status

    @property
    def transaction_type(self):
        if hasattr(self, '_transaction_type'):
            return self._transaction_type

    @transaction_type.setter
    def transaction_type(self, transaction_type):
        """

        :param transaction_type: The type of transaction that we are recording - e.g. Trade, Payment, Coupon
        :return:
        """
        if transaction_type not in TRANSACTION_TYPES:
            raise ValueError(ERROR_LOOKUP.get('transaction_type_invalid') % (transaction_type, self.transaction_id,
                                                                             self.asset_manager_id))
        else:
            self._transaction_type = transaction_type

    def charges_net_effect(self):
        """
        The total effect of the net_affecting charges (note affect vs effect here).

        Currently this is single currency only (AMAAS-110).

        Cast to Decimal in case the result is zero (no net_affecting charges).

        :return:
        """
        return Decimal(sum([charge.charge_value for charge in self.charges.values()
                            if charge.active and charge.net_affecting]))

    def charge_types(self):
        """
        TODO - are these helper functions useful?
        :return:
        """
        return self.charges.keys()

    def code_types(self):
        """
        TODO - are these helper functions useful?
        :return:
        """
        return self.codes.keys()

    def reference_types(self):
        """
        TODO - are these helper functions useful?
        :return:
        """
        return self.references.keys()

    def __str__(self):
        return "Transaction object - ID: %s" % self.transaction_id

    @property
    def postings(self):
        if hasattr(self, '_postings'):
            return self._postings
        else:
            raise TransactionNeedsSaving

    @postings.setter
    def postings(self, postings):
        """
        TODO - when do we save this from AMaaS Core?
        :param postings:
        :return:
        """
        if postings:
            self._postings = postings

    # Upsert methods for safely adding children - this is more important for cases where we trigger action when there
    # is a change, e.g. for the case of a @property on the collection.  Since we don't have that case yet for
    # transactions, I have not yet filled out all of these.
    def upsert_code(self, code_type, code):
        codes = copy.deepcopy(self.codes)
        codes.update({code_type: code})
        self.codes = codes

    def upsert_link_set(self, link_type, link_set):
        """
        Remove an item altogether by setting link_list to None.
        Currently, only links can contain multiple children of the same type.
        :param link_type:
        :param link_set:
        :return:
        """
        if link_set is None:
            self.links.pop(link_type, None)
            return
        links = copy.deepcopy(self.links)
        links.update({link_type: link_set})
        self.links = links

    def add_link(self, link_type, linked_transaction_id):
        new_link = Link(linked_transaction_id=linked_transaction_id)
        link_set = self.links.get(link_type)
        if link_set:
            if not isinstance(link_set, set):
                link_set = {link_set}
            link_set.add(new_link)
        else:
            link_set = new_link
        self.upsert_link_set(link_type=link_type, link_set=link_set)

    def remove_link(self, link_type, linked_transaction_id):
        link_set = self.links.get(link_type)
        if not link_set:
            raise KeyError(ERROR_LOOKUP.get('transaction_link_not_found'))
        if isinstance(link_set, Link):
            if link_set.linked_transaction_id == linked_transaction_id:
                link_set = None
            else:
                raise KeyError(ERROR_LOOKUP.get('transaction_link_not_found'))
        else:
            output = [link for link in link_set if link.linked_transaction_id == linked_transaction_id]
            if output:
                link_set.remove(output[0])
            else:
                raise KeyError(ERROR_LOOKUP.get('transaction_link_not_found'))
        self.upsert_link_set(link_type=link_type, link_set=link_set)



from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.core.amaas_model import AMaaSModel


class Address(AMaaSModel):

    def __init__(self, line_one, city, country_id, address_primary, line_two=None, region=None, postal_code=None,
                 active=True, version=1, *args, **kwargs):
        self.address_primary = address_primary
        self.line_one = line_one
        self.line_two = line_two
        self.city = city
        self.region = region
        self.postal_code = postal_code
        self.country_id = country_id
        self.active = active
        self.version = version
        super(Address, self).__init__(*args, **kwargs)


class Email(AMaaSModel):

    def __init__(self, email, email_primary, active=True, version=1, *args, **kwargs):
        self.email_primary = email_primary
        self.email = email
        self.active = active
        self.version = version
        super(Email, self).__init__(*args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from amaascore.config import ENDPOINTS
from amaascore.core.interface import Interface
from amaascore.parties.utils import json_to_party


class PartiesInterface(Interface):

    def __init__(self, logger=None):
        endpoint = ENDPOINTS.get('parties')
        self.logger = logger or logging.getLogger(__name__)
        super(PartiesInterface, self).__init__(endpoint=endpoint)

    def new(self, party):
        url = self.endpoint + '/parties'
        response = self.session.post(url, json=party.to_interface())
        if response.ok:
            party = json_to_party(response.json())
            return party
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def amend(self, party):
        url = '%s/parties/%s/%s' % (self.endpoint, party.asset_manager_id, party.party_id)
        response = self.session.put(url, json=party.to_interface())
        if response.ok:
            party = json_to_party(response.json())
            return party
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve(self, asset_manager_id, party_id):
        url = '%s/parties/%s/%s' % (self.endpoint, asset_manager_id, party_id)
        response = self.session.get(url)
        if response.ok:
            return json_to_party(response.json())
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def deactivate(self, asset_manager_id, party_id):
        url = '%s/parties/%s/%s' % (self.endpoint, asset_manager_id, party_id)
        json = {'party_status': 'Inactive'}
        response = self.session.patch(url, json=json)
        if response.ok:
            self.logger.info(response.text)
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def search(self, asset_manager_ids=None, party_ids=None):
        search_params = {}
        # Potentially roll this into a loop through args rather than explicitly named - depends on additional validation
        if asset_manager_ids:
            search_params['asset_manager_ids'] = asset_manager_ids
        if party_ids:
            search_params['party_ids'] = party_ids
        url = self.endpoint + '/parties'
        response = self.session.get(url, params=search_params)
        if response.ok:
            parties = [json_to_party(json_party) for json_party in response.json()]
            return parties
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def parties_by_asset_manager(self, asset_manager_id):
        url = '%s/parties/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url)
        if response.ok:
            parties = [json_to_party(json_party) for json_party in response.json()]
            return parties
        else:
            self.logger.error(response.text)
            response.raise_for_status()



from __future__ import absolute_import, division, print_function, unicode_literals

import csv

from amaascore.transactions.position import Position
from amaascore.transactions.transaction import Transaction


def json_to_position(json_position):
    position = Position(**json_position)
    return position


def json_to_transaction(json_transaction):
    for (collection_name, clazz) in Transaction.children().items():
        children = json_transaction.pop(collection_name, {})
        collection = {}
        for (child_type, child_json) in children.items():
            # Handle the case where there are multiple children for a given type - e.g. links
            if isinstance(child_json, list):
                child = set()
                for child_json_in_list in child_json:
                    child.add(clazz(**child_json_in_list))
            else:
                child = clazz(**child_json)
            collection[child_type] = child
        json_transaction[collection_name] = collection
    transaction = Transaction(**json_transaction)
    return transaction


def csv_filename_to_transactions(filename):
    with open(filename, 'r') as f:
        transactions = csv_stream_to_transactions(f)
    return transactions


def csv_stream_to_transactions(stream):
    reader = csv.DictReader(stream)
    transactions = []
    for row in reader:
        transactions.append(json_to_transaction(row))
    return transactions


def transactions_to_csv(transactions, filename):
    with open(filename, 'w') as csvfile:
        transactions_to_csv_stream(transactions=transactions, stream=csvfile)


def transactions_to_csv_stream(transactions, stream):
    if not transactions:
        return
    transaction_dicts = []
    for transaction in transactions:
        transaction_dict = transaction.to_json()
        # FOR NOW - remove all children
        [transaction_dict.pop(child, None) for child in Transaction.children().keys()]
        transaction_dicts.append(transaction_dict)
    fieldnames = transaction_dicts[0].keys()
    writer = csv.DictWriter(stream, fieldnames=fieldnames)
    writer.writeheader()
    writer.writerows(transaction_dicts)



from __future__ import absolute_import, division, print_function, unicode_literals

import random

from amaascore.core.reference import Reference
from amaascore.parties.broker import Broker
from amaascore.parties.children import Address, Email
from amaascore.parties.individual import Individual
from amaascore.parties.party import Party
from amaascore.tools.helpers import random_string


def generate_common(asset_manager_id, party_id, party_status):
    common = {'asset_manager_id': asset_manager_id or random.randint(1, 1000),
              'party_id': party_id or str(random.randint(1, 1000)),
              'party_status': party_status or 'Active'
              }

    return common


def generate_party(asset_manager_id=None, party_id=None, party_status=None):
    references = {'PartyDB': Reference(random_string(10))}
    attributes = generate_common(asset_manager_id=asset_manager_id, party_id=party_id, party_status=party_status)
    party = Party(**attributes)
    # Does this screw up the party due to mutability concerns?  Do some testing.
    party.references.update(references)
    party.upsert_address('Registered', generate_address(address_primary=True))
    party.upsert_email('Office', generate_email(email_primary=True))
    return party


def generate_broker(asset_manager_id=None, party_id=None, party_status=None):
    references = {'LEI': Reference(random_string(10))}
    attributes = generate_common(asset_manager_id=asset_manager_id, party_id=party_id, party_status=party_status)
    broker = Broker(**attributes)
    # Does this screw up the party due to mutability concerns?  Do some testing.
    broker.references.update(references)
    broker.upsert_address('Registered', generate_address(address_primary=True))
    broker.upsert_email('Office', generate_email(email_primary=True))
    return broker


def generate_individual(asset_manager_id=None, party_id=None, party_status=None):
    attributes = generate_common(asset_manager_id=asset_manager_id, party_id=party_id, party_status=party_status)
    individual = Individual(**attributes)
    return individual


def generate_address(country_id=None, address_primary=False):
    address = Address(line_one=random_string(20),
                      line_two=random.choice([None, random_string(10)]),
                      city=random_string(10),
                      region=random_string(10),
                      postal_code=random_string(6),
                      country_id=country_id or random_string(3),  # Make this a real country code
                      address_primary=address_primary)
    return address


def generate_email(email=None, email_primary=False):
    return Email(email=email or (random_string(10) + '@amaas.com'), email_primary=email_primary)



from __future__ import absolute_import, division, print_function, unicode_literals

from decimal import Decimal
import random
import string


def random_string(length):
    return ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(length))


def random_decimal():
    return Decimal(random.random() * 100).quantize(Decimal('0.01'))



from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
from decimal import Decimal
import random

from amaascore.assets.asset import Asset
from amaascore.assets.bond import BondGovernment
from amaascore.assets.bond_option import BondOption
from amaascore.assets.foreign_exchange import ForeignExchange
from amaascore.core.reference import Reference
from amaascore.tools.helpers import random_string

REFERENCE_TYPES = ['External']


def generate_common(asset_manager_id=None, asset_id=None):

    common = {'asset_manager_id': asset_manager_id or random.randint(1, 1000),
              'asset_id': asset_id or str(random.randint(1, 1000)),
              'currency': random.choice(['SGD', 'USD'])
              }

    return common


def generate_asset(asset_manager_id=None, asset_id=None, fungible=None):

    common = generate_common(asset_manager_id=asset_manager_id, asset_id=asset_id)
    common['fungible'] = random.choice([True, False]) if fungible is None else fungible

    asset = Asset(**common)
    references = {ref_type: Reference(reference_value=random_string(10)) for ref_type in REFERENCE_TYPES}

    asset.references.update(references)
    return asset


def generate_bond(asset_manager_id=None, asset_id=None):
    props = generate_common(asset_manager_id=asset_manager_id, asset_id=asset_id)
    bond = BondGovernment(par=Decimal('1000'),
                          pay_frequency='M',  # Need to check how we want to represent this
                          coupon=Decimal('0.05'),
                          **props)
    return bond


def generate_bond_option(asset_manager_id=None, asset_id=None, option_type=None, strike=None, option_style=None):
    props = generate_common(asset_manager_id=asset_manager_id, asset_id=asset_id)
    asset = BondOption(underlying_asset_id=random_string(10),
                       option_style=option_style or random.choice(['European', 'American']),
                       option_type=option_type or random.choice(['Put', 'Call']),
                       strike=strike or Decimal(random.uniform(99.0, 102.0)).quantize(Decimal('0.05')),
                       **props)
    return asset


def generate_foreignexchange(asset_id=None):
    asset = ForeignExchange(asset_id=asset_id)
    return asset

